
#include "postgres.h"
#include "fmgr.h"
#include "libpq/pqformat.h"
#include "access/gist.h"
#include "access/stratnum.h"
#include "utils/array.h"
#include "catalog/pg_type.h"
#include <limits.h>
#include <math.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

#if INT_MAX < 0x7fffffff
#error Expected int type to be at least 32 bit wide
#endif
#if LLONG_MAX < 0x7fffffffffffffff
#error Expected long long type to be at least 64 bit wide
#endif

#define CFL_TIMESTAMP_MAX (((int64_t)1e15)-1)
#define CFL_TIMESTAMP_MIN (-CFL_TIMESTAMP_MAX)

typedef struct {
  int64_t timestamp;
  int32_t actor;
} cfl_actortime;

typedef struct {
  int64_t context;
  int64_t timestamp;
  int32_t actor;
} cfl_ctxactortime;

typedef struct {
  char header[VARHDRSZ];
  int32_t nactors;
  int64_t context;
  int64_t timestamp;
  int32_t actors[FLEXIBLE_ARRAY_MEMBER];
} cfl_ctxactortime_query;

typedef struct {
  int32_t lower_actor;
  int32_t upper_actor;
  int64_t lower_timestamp;
  int64_t upper_timestamp;
} cfl_actortime_key;

typedef struct {
  int64_t lower_context;
  int64_t upper_context;
  int32_t lower_actor;
  int32_t upper_actor;
  int64_t lower_timestamp;
  int64_t upper_timestamp;
} cfl_ctxactortime_key;

#define CFL_HUGE_DISTANCE 1e38

static int cfl_compare_actor(const int32_t *a, const int32_t *b) {
  if (*a < *b) return -1;
  if (*a > *b) return 1;
  return 0;
}

static void cfl_finalize_ctxactortime_query(
  cfl_ctxactortime_query *query, int nactors
) {
  int i, j;
  qsort(query->actors, nactors, sizeof(int32_t), (void *)cfl_compare_actor);
  for (i=0, j=0; i<nactors; i++) {
    if (!i || query->actors[i] != query->actors[i-1]) {
      query->actors[j++] = query->actors[i];
    }
  }
  SET_VARSIZE(
    query,
    offsetof(cfl_ctxactortime_query, actors) + j * sizeof(int32_t)
  );
  query->nactors = j;
}

static int cfl_compare_actortime_key_ptr(
  const cfl_actortime_key **a, const cfl_actortime_key **b
) {
  if ((*a)->lower_actor < (*b)->lower_actor) return -1;
  if ((*a)->lower_actor > (*b)->lower_actor) return 1;
  if ((*a)->lower_timestamp < (*b)->lower_timestamp) return -1;
  if ((*a)->lower_timestamp > (*b)->lower_timestamp) return 1;
  if ((*a) < (*b)) return -1;
  if ((*a) > (*b)) return 1;
  return 0;
}

static int cfl_compare_ctxactortime_key_ptr(
  const cfl_ctxactortime_key **a, const cfl_ctxactortime_key **b
) {
  if ((*a)->lower_context < (*b)->lower_context) return -1;
  if ((*a)->lower_context > (*b)->lower_context) return 1;
  if ((*a)->lower_actor < (*b)->lower_actor) return -1;
  if ((*a)->lower_actor > (*b)->lower_actor) return 1;
  if ((*a)->lower_timestamp < (*b)->lower_timestamp) return -1;
  if ((*a)->lower_timestamp > (*b)->lower_timestamp) return 1;
  if ((*a) < (*b)) return -1;
  if ((*a) > (*b)) return 1;
  return 0;
}

static void cfl_actortime_key_union(
  cfl_actortime_key *dst, const cfl_actortime_key *src
) {
  if (src->lower_actor < dst->lower_actor) {
    dst->lower_actor = src->lower_actor;
    dst->lower_timestamp = src->lower_timestamp;
  } else if (
    src->lower_actor == dst->lower_actor &&
    src->lower_timestamp < dst->lower_timestamp
  ) {
    dst->lower_timestamp = src->lower_timestamp;
  }
  if (src->upper_actor > dst->upper_actor) {
    dst->upper_actor = src->upper_actor;
    dst->upper_timestamp = src->upper_timestamp;
  } else if (
    src->upper_actor == dst->upper_actor &&
    src->upper_timestamp > dst->upper_timestamp
  ) {
    dst->upper_timestamp = src->upper_timestamp;
  }
}

static void cfl_ctxactortime_key_union(
  cfl_ctxactortime_key *dst, const cfl_ctxactortime_key *src
) {
  if (src->lower_context < dst->lower_context) {
    dst->lower_context = src->lower_context;
    dst->lower_actor = src->lower_actor;
    dst->lower_timestamp = src->lower_timestamp;
  } else if (src->lower_context == dst->lower_context) {
    if (src->lower_actor < dst->lower_actor) {
      dst->lower_actor = src->lower_actor;
      dst->lower_timestamp = src->lower_timestamp;
    } else if (
      src->lower_actor == dst->lower_actor &&
      src->lower_timestamp < dst->lower_timestamp
    ) {
      dst->lower_timestamp = src->lower_timestamp;
    }
  }
  if (src->upper_context > dst->upper_context) {
    dst->upper_context = src->upper_context;
    dst->upper_actor = src->upper_actor;
    dst->upper_timestamp = src->upper_timestamp;
  } else if (src->upper_context == dst->upper_context) {
    if (src->upper_actor > dst->upper_actor) {
      dst->upper_actor = src->upper_actor;
      dst->upper_timestamp = src->upper_timestamp;
    } else if (
      src->upper_actor == dst->upper_actor &&
      src->upper_timestamp > dst->upper_timestamp
    ) {
      dst->upper_timestamp = src->upper_timestamp;
    }
  }
}

PG_FUNCTION_INFO_V1(cfl_notimpl);
Datum cfl_notimpl(PG_FUNCTION_ARGS) {
  ereport(ERROR, (errmsg("not implemented by conflux extension")));
}

PG_FUNCTION_INFO_V1(cfl_actortime_in);
Datum cfl_actortime_in(PG_FUNCTION_ARGS) {
  char *str = PG_GETARG_CSTRING(0);
  int actor;
  long long timestamp;
  int len;
  if (sscanf(str, "%d@%lld%n", &actor, &timestamp, &len) >= 2) {
    if (
      !str[len] &&
      actor >= INT32_MIN && actor <= INT32_MAX &&
      timestamp >= INT64_MIN && timestamp <= INT64_MAX
    ) {
      cfl_actortime *actortime;
      if (timestamp < CFL_TIMESTAMP_MIN || timestamp > CFL_TIMESTAMP_MAX) {
        ereport(ERROR, (
          errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
          errmsg(
            "numeric value out of range for type actortime: %lld",
            (long long)timestamp
          )
        ));
      }
      actortime = (cfl_actortime *)palloc(sizeof(cfl_actortime));
      actortime->actor = actor;
      actortime->timestamp = timestamp;
      PG_RETURN_POINTER(actortime);
    }
  }
  ereport(ERROR, (
    errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
    errmsg("invalid input syntax for type actortime: \"%s\"", str)
  ));
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_in);
Datum cfl_ctxactortime_in(PG_FUNCTION_ARGS) {
  char *str = PG_GETARG_CSTRING(0);
  long long context;
  int actor;
  long long timestamp;
  int len;
  if (sscanf(str, "%lld:%d@%lld%n", &context, &actor, &timestamp, &len) >= 3) {
    if (
      !str[len] &&
      context >= INT64_MIN && context <= INT64_MAX &&
      actor >= INT32_MIN && actor <= INT32_MAX &&
      timestamp >= INT64_MIN && timestamp <= INT64_MAX
    ) {
      cfl_ctxactortime *ctxactortime;
      if (timestamp < CFL_TIMESTAMP_MIN || timestamp > CFL_TIMESTAMP_MAX) {
        ereport(ERROR, (
          errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
          errmsg(
            "numeric value out of range for type ctxactortime: %lld",
            (long long)timestamp
          )
        ));
      }
      ctxactortime = (cfl_ctxactortime *)palloc(sizeof(cfl_ctxactortime));
      ctxactortime->context = context;
      ctxactortime->actor = actor;
      ctxactortime->timestamp = timestamp;
      PG_RETURN_POINTER(ctxactortime);
    }
  }
  ereport(ERROR, (
    errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
    errmsg("invalid input syntax for type actortime: \"%s\"", str)
  ));
}

PG_FUNCTION_INFO_V1(cfl_actortime_out);
Datum cfl_actortime_out(PG_FUNCTION_ARGS) {
  cfl_actortime *actortime = (cfl_actortime *)PG_GETARG_POINTER(0);
  PG_RETURN_CSTRING(psprintf(
    "%d@%lld",
    (int)actortime->actor,
    (long long)actortime->timestamp
  ));
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_out);
Datum cfl_ctxactortime_out(PG_FUNCTION_ARGS) {
  cfl_ctxactortime *ctxactortime = (cfl_ctxactortime *)PG_GETARG_POINTER(0);
  PG_RETURN_CSTRING(psprintf(
    "%lld:%d@%lld",
    (long long)ctxactortime->context,
    (int)ctxactortime->actor,
    (long long)ctxactortime->timestamp
  ));
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_query_in);
Datum cfl_ctxactortime_query_in(PG_FUNCTION_ARGS) {
  char *str = PG_GETARG_CSTRING(0);
  char *strptr;
  int nactors = 1, i;
  long long scanval;
  int len;
  cfl_ctxactortime_query *query;
  for (strptr=str; *strptr; strptr++) {
    if (*strptr == ',') {
      nactors++;
      if (nactors == INT_MAX || nactors == INT32_MAX) {
        goto cfl_ctxactortime_query_in_error;
      }
    }
  }
  if (
    nactors >
    (SIZE_MAX - offsetof(cfl_ctxactortime_query, actors)) / sizeof(int32_t)
  ) {
    goto cfl_ctxactortime_query_in_error;
  }
  query = (cfl_ctxactortime_query *)palloc(
    offsetof(cfl_ctxactortime_query, actors) + nactors * sizeof(int32_t)
  );
  strptr = str;
  if (
    sscanf(strptr, "%lld:%n", &scanval, &len) < 1 ||
    scanval < INT64_MIN || scanval > INT64_MAX
  ) {
    goto cfl_ctxactortime_query_in_error;
  }
  strptr += len;
  query->context = scanval;
  if (*strptr == '@') nactors = 0;
  for (i=0; i<nactors; i++) {
    if (
      sscanf(strptr, i ? ",%lld%n" : "%lld%n", &scanval, &len) < 1 ||
      scanval < INT32_MIN || scanval > INT32_MAX
    ) {
      goto cfl_ctxactortime_query_in_error;
    }
    strptr += len;
    query->actors[i] = scanval;
  }
  if (
    sscanf(strptr, "@%lld%n", &scanval, &len) != 1 ||
    scanval < INT64_MIN || scanval > INT64_MAX
  ) {
    goto cfl_ctxactortime_query_in_error;
  }
  strptr += len;
  if (scanval > CFL_TIMESTAMP_MAX+1) scanval = CFL_TIMESTAMP_MAX+1;
  query->timestamp = scanval;
  if (*strptr) goto cfl_ctxactortime_query_in_error;
  cfl_finalize_ctxactortime_query(query, nactors);
  PG_RETURN_POINTER(query);
  cfl_ctxactortime_query_in_error:
  ereport(ERROR, (
    errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
    errmsg("invalid input syntax for type ctxactortime_query: \"%s\"", str)
  ));
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_query_out);
Datum cfl_ctxactortime_query_out(PG_FUNCTION_ARGS) {
  cfl_ctxactortime_query *query = (cfl_ctxactortime_query *)PG_DETOAST_DATUM(
    PG_GETARG_POINTER(0)
  );
  StringInfoData buf;
  int i;
  initStringInfo(&buf);
  appendStringInfo(&buf, "%lld:", (long long)query->context);
  for (i=0; i<query->nactors; i++) {
    appendStringInfo(&buf, i ? ",%d" : "%d", (int)query->actors[i]);
  }
  appendStringInfo(&buf, "@%lld", (long long)query->timestamp);
  PG_FREE_IF_COPY(query, 0);
  PG_RETURN_CSTRING(buf.data);
}

PG_FUNCTION_INFO_V1(cfl_actortime_recv);
Datum cfl_actortime_recv(PG_FUNCTION_ARGS) {
  StringInfo buf = (StringInfo)PG_GETARG_POINTER(0);
  cfl_actortime *actortime = (cfl_actortime *)palloc(sizeof(cfl_actortime));
  actortime->actor = pq_getmsgint(buf, 4);
  actortime->timestamp = pq_getmsgint64(buf);
  PG_RETURN_POINTER(actortime);
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_recv);
Datum cfl_ctxactortime_recv(PG_FUNCTION_ARGS) {
  StringInfo buf = (StringInfo)PG_GETARG_POINTER(0);
  cfl_ctxactortime *ctxactortime =
    (cfl_ctxactortime *)palloc(sizeof(cfl_ctxactortime));
  ctxactortime->context = pq_getmsgint64(buf);
  ctxactortime->actor = pq_getmsgint(buf, 4);
  ctxactortime->timestamp = pq_getmsgint64(buf);
  PG_RETURN_POINTER(ctxactortime);
}

PG_FUNCTION_INFO_V1(cfl_actortime_send);
Datum cfl_actortime_send(PG_FUNCTION_ARGS) {
  cfl_actortime *actortime = (cfl_actortime *)PG_GETARG_POINTER(0);
  StringInfoData buf;
  pq_begintypsend(&buf);
  pq_sendint(&buf, actortime->actor, 4);
  pq_sendint64(&buf, actortime->timestamp);
  PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_send);
Datum cfl_ctxactortime_send(PG_FUNCTION_ARGS) {
  cfl_ctxactortime *ctxactortime = (cfl_ctxactortime *)PG_GETARG_POINTER(0);
  StringInfoData buf;
  pq_begintypsend(&buf);
  pq_sendint64(&buf, ctxactortime->context);
  pq_sendint(&buf, ctxactortime->actor, 4);
  pq_sendint64(&buf, ctxactortime->timestamp);
  PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

PG_FUNCTION_INFO_V1(cfl_actortime_new);
Datum cfl_actortime_new(PG_FUNCTION_ARGS) {
  int32_t actor = PG_GETARG_INT32(0);
  int64_t timestamp = PG_GETARG_INT64(1);
  cfl_actortime *actortime;
  if (timestamp < CFL_TIMESTAMP_MIN || timestamp > CFL_TIMESTAMP_MAX) {
    ereport(ERROR, (
      errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
      errmsg(
        "numeric value out of range for type actortime: %lld",
        (long long)timestamp
      )
    ));
  }
  actortime = palloc(sizeof(cfl_actortime));
  actortime->actor = actor;
  actortime->timestamp = timestamp;
  PG_RETURN_POINTER(actortime);
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_new);
Datum cfl_ctxactortime_new(PG_FUNCTION_ARGS) {
  int64_t context = PG_GETARG_INT64(0);
  int32_t actor = PG_GETARG_INT32(1);
  int64_t timestamp = PG_GETARG_INT64(2);
  cfl_ctxactortime *ctxactortime;
  if (timestamp < CFL_TIMESTAMP_MIN || timestamp > CFL_TIMESTAMP_MAX) {
    ereport(ERROR, (
      errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
      errmsg(
        "numeric value out of range for type ctxactortime: %lld",
        (long long)timestamp
      )
    ));
  }
  ctxactortime = palloc(sizeof(cfl_ctxactortime));
  ctxactortime->context = context;
  ctxactortime->actor = actor;
  ctxactortime->timestamp = timestamp;
  PG_RETURN_POINTER(ctxactortime);
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_query_new);
Datum cfl_ctxactortime_query_new(PG_FUNCTION_ARGS) {
  ArrayType *array;
  Datum *actors;
  bool *nulls;
  int nactors, i, j;
  cfl_ctxactortime_query *query;
  if (PG_ARGISNULL(0) || PG_ARGISNULL(1)) PG_RETURN_NULL();
  array = PG_GETARG_ARRAYTYPE_P(1);
  deconstruct_array(array, INT4OID, 4, true, 'i', &actors, &nulls, &nactors);
  if (
    nactors >= INT_MAX || nactors >= INT32_MAX ||
    nactors >
    (SIZE_MAX - offsetof(cfl_ctxactortime_query, actors)) / sizeof(int32_t)
  ) {
    ereport(ERROR, (
      errcode(ERRCODE_INVALID_PARAMETER_VALUE),
      errmsg("array passed to ctxactortime_query constructor is too long")
    ));
  }
  for (i=0, j=0; i<nactors; i++) { if (!nulls[i]) j++; }
  query = (cfl_ctxactortime_query *)palloc(
    offsetof(cfl_ctxactortime_query, actors) + j * sizeof(int32_t)
  );
  query->context = PG_GETARG_INT64(0);
  if (PG_ARGISNULL(2)) query->timestamp = CFL_TIMESTAMP_MAX+1;
  else {
    int64_t timestamp = PG_GETARG_INT64(2);
    if (timestamp > CFL_TIMESTAMP_MAX+1) timestamp = CFL_TIMESTAMP_MAX+1;
    query->timestamp = timestamp;
  }
  for (i=0, j=0; i<nactors; i++) {
    if (!nulls[i]) query->actors[j++] = DatumGetInt64(actors[i]);
  }
  pfree(actors);
  pfree(nulls);
  cfl_finalize_ctxactortime_query(query, j);
  PG_FREE_IF_COPY(array, 1);
  PG_RETURN_POINTER(query);
}

PG_FUNCTION_INFO_V1(cfl_actortime_match);
Datum cfl_actortime_match(PG_FUNCTION_ARGS) {
  cfl_actortime *actortime = (cfl_actortime *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  if (query->timestamp <= actortime->timestamp) PG_RETURN_BOOL(false);
  if (bsearch(
    &actortime->actor,
    query->actors,
    query->nactors,
    sizeof(int32_t),
    (void *)cfl_compare_actor
  )) {
    PG_FREE_IF_COPY(query, 1);
    PG_RETURN_BOOL(true);
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_match);
Datum cfl_ctxactortime_match(PG_FUNCTION_ARGS) {
  cfl_ctxactortime *ctxactortime = (cfl_ctxactortime *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  if (
    query->context != ctxactortime->context ||
    query->timestamp <= ctxactortime->timestamp
  ) PG_RETURN_BOOL(false);
  if (bsearch(
    &ctxactortime->actor,
    query->actors,
    query->nactors,
    sizeof(int32_t),
    (void *)cfl_compare_actor
  )) {
    PG_FREE_IF_COPY(query, 1);
    PG_RETURN_BOOL(true);
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(cfl_actortime_distance);
Datum cfl_actortime_distance(PG_FUNCTION_ARGS) {
  cfl_actortime *actortime = (cfl_actortime *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  if (query->timestamp <= actortime->timestamp) PG_RETURN_FLOAT8(INFINITY);
  if (bsearch(
    &actortime->actor,
    query->actors,
    query->nactors,
    sizeof(int32_t),
    (void *)cfl_compare_actor
  )) {
    PG_FREE_IF_COPY(query, 1);
    PG_RETURN_FLOAT8(query->timestamp - actortime->timestamp);
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_FLOAT8(INFINITY);
}

PG_FUNCTION_INFO_V1(cfl_ctxactortime_distance);
Datum cfl_ctxactortime_distance(PG_FUNCTION_ARGS) {
  cfl_ctxactortime *ctxactortime = (cfl_ctxactortime *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  if (
    query->context != ctxactortime->context ||
    query->timestamp <= ctxactortime->timestamp
  ) PG_RETURN_FLOAT8(INFINITY);
  if (bsearch(
    &ctxactortime->actor,
    query->actors,
    query->nactors,
    sizeof(int32_t),
    (void *)cfl_compare_actor
  )) {
    PG_FREE_IF_COPY(query, 1);
    PG_RETURN_FLOAT8(query->timestamp - ctxactortime->timestamp);
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_FLOAT8(INFINITY);
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_consistent);
Datum cfl_gist_actortime_consistent(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  cfl_actortime_key *key = (cfl_actortime_key *)DatumGetPointer(entry->key);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  bool *recheck = (bool *)PG_GETARG_POINTER(4);
  if (query->nactors) {
    int i = 0, j = query->nactors-1, k;
    while (i != j) {
      k = (i+j) / 2;
      if (query->actors[k] < key->lower_actor) i = k + 1;
      else j = k;
    }
    if (
      query->actors[i] >= key->lower_actor &&
      query->actors[i] <= key->upper_actor
    ) {
      if (key->lower_actor == key->upper_actor) {
        if (key->lower_timestamp < query->timestamp) {
          *recheck = false;
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_BOOL(true);
        } else {
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_BOOL(false);
        }
      } else {
        *recheck = true;
        PG_FREE_IF_COPY(query, 1);
        PG_RETURN_BOOL(true);
      }
    }
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_consistent);
Datum cfl_gist_ctxactortime_consistent(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_key *key =
    (cfl_ctxactortime_key *)DatumGetPointer(entry->key);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  bool *recheck = (bool *)PG_GETARG_POINTER(4);
  if (
    query->context >= key->lower_context &&
    query->context <= key->upper_context &&
    query->nactors
  ) {
    if (key->lower_context == key->upper_context) {
      int i = 0, j = query->nactors-1, k;
      while (i != j) {
        k = (i+j) / 2;
        if (query->actors[k] < key->lower_actor) i = k + 1;
        else j = k;
      }
      if (
        query->actors[i] >= key->lower_actor &&
        query->actors[i] <= key->upper_actor
      ) {
        if (key->lower_actor == key->upper_actor) {
          if (key->lower_timestamp < query->timestamp) {
            *recheck = false;
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_BOOL(true);
          } else {
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_BOOL(false);
          }
        } else {
          *recheck = true;
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_BOOL(true);
        }
      }
    } else {
      *recheck = true;
      PG_FREE_IF_COPY(query, 1);
      PG_RETURN_BOOL(true);
    }
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_BOOL(false);
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_union);
Datum cfl_gist_actortime_union(PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *)PG_GETARG_POINTER(0);
  cfl_actortime_key *key = palloc(sizeof(cfl_actortime_key));
  int i;
  *key = *(cfl_actortime_key *)DatumGetPointer(entryvec->vector[0].key);
  for (i=1; i<entryvec->n; i++) {
    cfl_actortime_key_union(
      key,
      (cfl_actortime_key *)DatumGetPointer(entryvec->vector[i].key)
    );
  }
  PG_RETURN_POINTER(key);
}

PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_union);
Datum cfl_gist_ctxactortime_union(PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_key *key = palloc(sizeof(cfl_ctxactortime_key));
  int i;
  *key = *(cfl_ctxactortime_key *)DatumGetPointer(entryvec->vector[0].key);
  for (i=1; i<entryvec->n; i++) {
    cfl_ctxactortime_key_union(
      key,
      (cfl_ctxactortime_key *)DatumGetPointer(entryvec->vector[i].key)
    );
  }
  PG_RETURN_POINTER(key);
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_compress);
Datum cfl_gist_actortime_compress(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  GISTENTRY *retval;
  if (entry->leafkey) {
    cfl_actortime *actortime = (cfl_actortime *)DatumGetPointer(entry->key);
    cfl_actortime_key *key = palloc(sizeof(cfl_actortime_key));
    key->lower_actor = actortime->actor;
    key->upper_actor = actortime->actor;
    key->lower_timestamp = actortime->timestamp;
    key->upper_timestamp = actortime->timestamp;
    retval = palloc(sizeof(GISTENTRY));
    gistentryinit(
      *retval, PointerGetDatum(key),
      entry->rel, entry->page, entry->offset, false
    );
  } else {
    retval = entry;
  }
  PG_RETURN_POINTER(retval);
}

PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_compress);
Datum cfl_gist_ctxactortime_compress(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  GISTENTRY *retval;
  if (entry->leafkey) {
    cfl_ctxactortime *ctxactortime =
      (cfl_ctxactortime *)DatumGetPointer(entry->key);
    cfl_ctxactortime_key *key = palloc(sizeof(cfl_ctxactortime_key));
    key->lower_context = ctxactortime->context;
    key->upper_context = ctxactortime->context;
    key->lower_actor = ctxactortime->actor;
    key->upper_actor = ctxactortime->actor;
    key->lower_timestamp = ctxactortime->timestamp;
    key->upper_timestamp = ctxactortime->timestamp;
    retval = palloc(sizeof(GISTENTRY));
    gistentryinit(
      *retval, PointerGetDatum(key),
      entry->rel, entry->page, entry->offset, false
    );
  } else {
    retval = entry;
  }
  PG_RETURN_POINTER(retval);
}

PG_FUNCTION_INFO_V1(cfl_gist_decompress);
Datum cfl_gist_decompress(PG_FUNCTION_ARGS) {
  PG_RETURN_POINTER(PG_GETARG_POINTER(0));
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_penalty);
Datum cfl_gist_actortime_penalty(PG_FUNCTION_ARGS) {
  GISTENTRY *origentry = (GISTENTRY *)PG_GETARG_POINTER(0);
  GISTENTRY *newentry = (GISTENTRY *)PG_GETARG_POINTER(1);
  float *penalty = (float *)PG_GETARG_POINTER(2);
  cfl_actortime_key *orig =
    (cfl_actortime_key *)DatumGetPointer(origentry->key);
  cfl_actortime_key grown = *orig;
  cfl_actortime_key_union(
    &grown,
    (cfl_actortime_key *)DatumGetPointer(newentry->key)
  );
  if (
    grown.lower_actor == orig->lower_actor &&
    grown.upper_actor == orig->upper_actor
  ) {
    if (
      grown.lower_timestamp == orig->lower_timestamp &&
      grown.upper_timestamp == orig->upper_timestamp
    ) {
      *penalty = 0.0;
    } else {
      *penalty = 1e-20 * (uint64_t)(
        (grown.upper_timestamp - grown.lower_timestamp) -
        (orig->upper_timestamp - orig->lower_timestamp)
      );
    }
  } else {
    *penalty = (uint32_t)(
      (grown.upper_actor - grown.lower_actor) -
      (orig->upper_actor - orig->lower_actor)
    );
  }
  PG_RETURN_POINTER(penalty);
}

PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_penalty);
Datum cfl_gist_ctxactortime_penalty(PG_FUNCTION_ARGS) {
  GISTENTRY *origentry = (GISTENTRY *)PG_GETARG_POINTER(0);
  GISTENTRY *newentry = (GISTENTRY *)PG_GETARG_POINTER(1);
  float *penalty = (float *)PG_GETARG_POINTER(2);
  cfl_ctxactortime_key *orig =
    (cfl_ctxactortime_key *)DatumGetPointer(origentry->key);
  cfl_ctxactortime_key grown = *orig;
  cfl_ctxactortime_key_union(
    &grown,
    (cfl_ctxactortime_key *)DatumGetPointer(newentry->key)
  );
  if (
    grown.lower_context == orig->lower_context &&
    grown.upper_context == orig->upper_context
  ) {
    if (
      grown.lower_actor == orig->lower_actor &&
      grown.upper_actor == orig->upper_actor
    ) {
      if (
        grown.lower_timestamp == orig->lower_timestamp &&
        grown.upper_timestamp == orig->upper_timestamp
      ) {
        *penalty = 0.0;
      } else {
        *penalty = 1e-20 * (uint64_t)(
          (grown.upper_timestamp - grown.lower_timestamp) -
          (orig->upper_timestamp - orig->lower_timestamp)
        );
      }
    } else {
      *penalty = (uint32_t)(
        (grown.upper_actor - grown.lower_actor) -
        (orig->upper_actor - orig->lower_actor)
      );
    }
  } else {
    *penalty = 1e20 * (uint64_t)(
      (grown.upper_context - grown.lower_context) -
      (orig->upper_context - orig->lower_context)
    );
  }
  PG_RETURN_POINTER(penalty);
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_picksplit);
Datum cfl_gist_actortime_picksplit(PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *)PG_GETARG_POINTER(0);
  GIST_SPLITVEC *v = (GIST_SPLITVEC *)PG_GETARG_POINTER(1);
  OffsetNumber i;
  int nkeys = entryvec->n - FirstOffsetNumber;
  const cfl_actortime_key **keys = palloc(nkeys * sizeof(cfl_actortime_key *));
  const cfl_actortime_key *key;
  const cfl_actortime_key *middle;
  cfl_actortime_key *unionL = NULL;
  cfl_actortime_key *unionR = NULL;
  for (i=FirstOffsetNumber; i<entryvec->n; i=OffsetNumberNext(i)) {
    keys[i-FirstOffsetNumber] =
      (cfl_actortime_key *)DatumGetPointer(entryvec->vector[i].key);
  }
  qsort(
    keys,
    nkeys,
    sizeof(cfl_actortime_key *),
    (void *)cfl_compare_actortime_key_ptr
  );
  middle = keys[nkeys/2];
  v->spl_left = (OffsetNumber *)palloc(entryvec->n * sizeof(OffsetNumber));
  v->spl_nleft = 0;
  v->spl_right = (OffsetNumber *)palloc(entryvec->n * sizeof(OffsetNumber));
  v->spl_nright = 0;
  for (i=FirstOffsetNumber; i<entryvec->n; i=OffsetNumberNext(i)) {
    key = (cfl_actortime_key *)DatumGetPointer(entryvec->vector[i].key);
    if (cfl_compare_actortime_key_ptr(&key, &middle) <= 0) {
      if (unionL) {
        cfl_actortime_key_union(unionL, key);
      } else {
        unionL = palloc(sizeof(cfl_actortime_key));
        *unionL = *key;
      }
      v->spl_left[v->spl_nleft++] = i;
    } else {
      if (unionR) {
        cfl_actortime_key_union(unionR, key);
      } else {
        unionR = palloc(sizeof(cfl_actortime_key));
        *unionR = *key;
      }
      v->spl_right[v->spl_nright++] = i;
    }
  }
  v->spl_ldatum = PointerGetDatum(unionL);
  v->spl_rdatum = PointerGetDatum(unionR);
  PG_RETURN_POINTER(v);
}

PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_picksplit);
Datum cfl_gist_ctxactortime_picksplit(PG_FUNCTION_ARGS) {
  GistEntryVector *entryvec = (GistEntryVector *)PG_GETARG_POINTER(0);
  GIST_SPLITVEC *v = (GIST_SPLITVEC *)PG_GETARG_POINTER(1);
  OffsetNumber i;
  int nkeys = entryvec->n - FirstOffsetNumber;
  const cfl_ctxactortime_key **keys =
    palloc(nkeys * sizeof(cfl_ctxactortime_key *));
  const cfl_ctxactortime_key *key;
  const cfl_ctxactortime_key *middle;
  cfl_ctxactortime_key *unionL = NULL;
  cfl_ctxactortime_key *unionR = NULL;
  for (i=FirstOffsetNumber; i<entryvec->n; i=OffsetNumberNext(i)) {
    keys[i-FirstOffsetNumber] =
      (cfl_ctxactortime_key *)DatumGetPointer(entryvec->vector[i].key);
  }
  qsort(
    keys,
    nkeys,
    sizeof(cfl_ctxactortime_key *),
    (void *)cfl_compare_ctxactortime_key_ptr
  );
  middle = keys[nkeys/2];
  v->spl_left = (OffsetNumber *)palloc(entryvec->n * sizeof(OffsetNumber));
  v->spl_nleft = 0;
  v->spl_right = (OffsetNumber *)palloc(entryvec->n * sizeof(OffsetNumber));
  v->spl_nright = 0;
  for (i=FirstOffsetNumber; i<entryvec->n; i=OffsetNumberNext(i)) {
    key = (cfl_ctxactortime_key *)DatumGetPointer(entryvec->vector[i].key);
    if (cfl_compare_ctxactortime_key_ptr(&key, &middle) <= 0) {
      if (unionL) {
        cfl_ctxactortime_key_union(unionL, key);
      } else {
        unionL = palloc(sizeof(cfl_ctxactortime_key));
        *unionL = *key;
      }
      v->spl_left[v->spl_nleft++] = i;
    } else {
      if (unionR) {
        cfl_ctxactortime_key_union(unionR, key);
      } else {
        unionR = palloc(sizeof(cfl_ctxactortime_key));
        *unionR = *key;
      }
      v->spl_right[v->spl_nright++] = i;
    }
  }
  v->spl_ldatum = PointerGetDatum(unionL);
  v->spl_rdatum = PointerGetDatum(unionR);
  PG_RETURN_POINTER(v);
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_same);
Datum cfl_gist_actortime_same(PG_FUNCTION_ARGS) {
  cfl_actortime_key *key1 = (cfl_actortime_key *)PG_GETARG_POINTER(0);
  cfl_actortime_key *key2 = (cfl_actortime_key *)PG_GETARG_POINTER(1);
  bool *boolptr = (bool *)PG_GETARG_POINTER(2);
  *boolptr = (
    key1->lower_actor == key2->lower_actor &&
    key1->upper_actor == key2->upper_actor &&
    key1->lower_timestamp == key2->lower_timestamp &&
    key1->upper_timestamp == key2->upper_timestamp
  );
  PG_RETURN_POINTER(boolptr);
}

PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_same);
Datum cfl_gist_ctxactortime_same(PG_FUNCTION_ARGS) {
  cfl_ctxactortime_key *key1 = (cfl_ctxactortime_key *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_key *key2 = (cfl_ctxactortime_key *)PG_GETARG_POINTER(1);
  bool *boolptr = (bool *)PG_GETARG_POINTER(2);
  *boolptr = (
    key1->lower_context == key2->lower_context &&
    key1->upper_context == key2->upper_context &&
    key1->lower_actor == key2->lower_actor &&
    key1->upper_actor == key2->upper_actor &&
    key1->lower_timestamp == key2->lower_timestamp &&
    key1->upper_timestamp == key2->upper_timestamp
  );
  PG_RETURN_POINTER(boolptr);
}

PG_FUNCTION_INFO_V1(cfl_gist_actortime_distance);
Datum cfl_gist_actortime_distance(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  cfl_actortime_key *key = (cfl_actortime_key *)DatumGetPointer(entry->key);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  if (query->nactors) {
    int i = 0, j = query->nactors-1, k;
    while (i != j) {
      k = (i+j) / 2;
      if (query->actors[k] < key->lower_actor) i = k + 1;
      else j = k;
    }
    if (
      query->actors[i] >= key->lower_actor &&
      query->actors[i] <= key->upper_actor
    ) {
      if (key->lower_actor == key->upper_actor) {
        if (key->lower_timestamp < query->timestamp) {
          if (key->upper_timestamp >= query->timestamp) {
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_FLOAT8(0);
          } else {
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_FLOAT8(query->timestamp - key->upper_timestamp);
          }
        } else {
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_FLOAT8(CFL_HUGE_DISTANCE);
        }
      } else {
        PG_FREE_IF_COPY(query, 1);
        PG_RETURN_FLOAT8(0);
      }
    }
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_FLOAT8(CFL_HUGE_DISTANCE);
}
PG_FUNCTION_INFO_V1(cfl_gist_ctxactortime_distance);
Datum cfl_gist_ctxactortime_distance(PG_FUNCTION_ARGS) {
  GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);
  cfl_ctxactortime_key *key =
    (cfl_ctxactortime_key *)DatumGetPointer(entry->key);
  cfl_ctxactortime_query *query =
    (cfl_ctxactortime_query *)PG_DETOAST_DATUM(PG_GETARG_POINTER(1));
  if (
    query->context >= key->lower_context &&
    query->context <= key->upper_context &&
    query->nactors
  ) {
    if (key->lower_context == key->upper_context) {
      int i = 0, j = query->nactors-1, k;
      while (i != j) {
        k = (i+j) / 2;
        if (query->actors[k] < key->lower_actor) i = k + 1;
        else j = k;
      }
      if (
        query->actors[i] >= key->lower_actor &&
        query->actors[i] <= key->upper_actor
      ) {
        if (key->lower_actor == key->upper_actor) {
          if (key->lower_timestamp < query->timestamp) {
            if (key->upper_timestamp >= query->timestamp) {
              PG_FREE_IF_COPY(query, 1);
              PG_RETURN_FLOAT8(0);
            } else {
              PG_FREE_IF_COPY(query, 1);
              PG_RETURN_FLOAT8(query->timestamp - key->upper_timestamp);
            }
          } else {
            PG_FREE_IF_COPY(query, 1);
            PG_RETURN_FLOAT8(CFL_HUGE_DISTANCE);
          }
        } else {
          PG_FREE_IF_COPY(query, 1);
          PG_RETURN_FLOAT8(0);
        }
      }
    } else {
      PG_FREE_IF_COPY(query, 1);
      PG_RETURN_FLOAT8(0);
    }
  }
  PG_FREE_IF_COPY(query, 1);
  PG_RETURN_FLOAT8(CFL_HUGE_DISTANCE);
}

